@@ -25,11 +25,16 @@ import time
 
 from keystoneauth1.access import service_catalog as keystone_sc
 from keystoneauth1 import exceptions as keystone_exc
+from keystoneauth1 import identity as ksa_identity
+from keystoneauth1 import session as ksa_session
+from keystoneauth1 import token_endpoint as ksa_token_endpoint
 from oslo_concurrency import processutils
 from oslo_config import cfg
 from oslo_utils import units
 
 from glance_store import capabilities
+from glance_store.common import attachment_state_manager
+from glance_store.common import cinder_utils
 from glance_store.common import utils
 import glance_store.driver
 from glance_store import exceptions
@@ -37,6 +42,7 @@ from glance_store.i18n import _, _LE, _LI, _LW
 import glance_store.location
 
 try:
+    from cinderclient import api_versions
     from cinderclient import exceptions as cinder_exception
     from cinderclient.v3 import client as cinderclient
     from os_brick.initiator import connector
@@ -79,6 +85,8 @@ Related options:
     * cinder_store_user_name
     * cinder_store_project_name
     * cinder_store_password
+    * cinder_store_project_domain_name
+    * cinder_store_user_domain_name
 
 """),
     cfg.StrOpt('cinder_endpoint_template',
@@ -104,6 +112,8 @@ Related options:
     * cinder_store_user_name
     * cinder_store_project_name
     * cinder_store_password
+    * cinder_store_project_domain_name
+    * cinder_store_user_domain_name
     * cinder_catalog_info
 
 """),
@@ -215,6 +225,8 @@ Related options:
     * cinder_store_user_name
     * cinder_store_password
     * cinder_store_project_name
+    * cinder_store_project_domain_name
+    * cinder_store_user_domain_name
 
 """),
     cfg.StrOpt('cinder_store_user_name',
@@ -222,8 +234,9 @@ Related options:
                help="""
 User name to authenticate against cinder.
 
-This must be used with all the following related options. If any of these are
-not specified, the user of the current context is used.
+This must be used with all the following non-domain-related options.
+If any of these are not specified (except domain-related options),
+the user of the current context is used.
 
 Possible values:
     * A valid user name
@@ -232,14 +245,33 @@ Related options:
     * cinder_store_auth_address
     * cinder_store_password
     * cinder_store_project_name
+    * cinder_store_project_domain_name
+    * cinder_store_user_domain_name
+
+"""),
+    cfg.StrOpt('cinder_store_user_domain_name',
+               default='Default',
+               help="""
+Domain of the user to authenticate against cinder.
+
+Possible values:
+    * A valid domain name for the user specified by ``cinder_store_user_name``
+
+Related options:
+    * cinder_store_auth_address
+    * cinder_store_password
+    * cinder_store_project_name
+    * cinder_store_project_domain_name
+    * cinder_store_user_name
 
 """),
     cfg.StrOpt('cinder_store_password', secret=True,
                help="""
 Password for the user authenticating against cinder.
 
-This must be used with all the following related options. If any of these are
-not specified, the user of the current context is used.
+This must be used with all the following related options.
+If any of these are not specified (except domain-related options),
+the user of the current context is used.
 
 Possible values:
     * A valid password for the user specified by ``cinder_store_user_name``
@@ -248,6 +280,8 @@ Related options:
     * cinder_store_auth_address
     * cinder_store_user_name
     * cinder_store_project_name
+    * cinder_store_project_domain_name
+    * cinder_store_user_domain_name
 
 """),
     cfg.StrOpt('cinder_store_project_name',
@@ -258,8 +292,9 @@ Project name where the image volume is stored in cinder.
 If this configuration option is not set, the project in current context is
 used.
 
-This must be used with all the following related options. If any of these are
-not specified, the project of the current context is used.
+This must be used with all the following related options.
+If any of these are not specified (except domain-related options),
+the user of the current context is used.
 
 Possible values:
     * A valid project name
@@ -268,6 +303,25 @@ Related options:
     * ``cinder_store_auth_address``
     * ``cinder_store_user_name``
     * ``cinder_store_password``
+    * ``cinder_store_project_domain_name``
+    * ``cinder_store_user_domain_name``
+
+"""),
+    cfg.StrOpt('cinder_store_project_domain_name',
+               default='Default',
+               help="""
+Domain of the project where the image volume is stored in cinder.
+
+Possible values:
+    * A valid domain name of the project specified by
+      ``cinder_store_project_name``
+
+Related options:
+    * ``cinder_store_auth_address``
+    * ``cinder_store_user_name``
+    * ``cinder_store_password``
+    * ``cinder_store_project_domain_name``
+    * ``cinder_store_user_domain_name``
 
 """),
     cfg.StrOpt('rootwrap_config',
@@ -350,6 +404,34 @@ Possible values:
 """),
 ]
 
+CINDER_SESSION = None
+
+
+def _reset_cinder_session():
+    global CINDER_SESSION
+    CINDER_SESSION = None
+
+
+def get_cinder_session(conf):
+    global CINDER_SESSION
+    if not CINDER_SESSION:
+        auth = ksa_identity.V3Password(
+            password=conf.cinder_store_password,
+            username=conf.cinder_store_user_name,
+            user_domain_name=conf.cinder_store_user_domain_name,
+            project_name=conf.cinder_store_project_name,
+            project_domain_name=conf.cinder_store_project_domain_name,
+            auth_url=conf.cinder_store_auth_address
+        )
+        if conf.cinder_api_insecure:
+            verify = False
+        elif conf.cinder_ca_certificates_file:
+            verify = conf.cinder_ca_certificates_file
+        else:
+            verify = True
+        CINDER_SESSION = ksa_session.Session(auth=auth, verify=verify)
+    return CINDER_SESSION
+
 
 class StoreLocation(glance_store.location.StoreLocation):
 
@@ -397,6 +479,7 @@ class Store(glance_store.driver.Store):
             self.store_conf = getattr(self.conf, self.backend_group)
         else:
             self.store_conf = self.conf.glance_store
+        self.volume_api = cinder_utils.API()
 
     def _set_url_prefix(self):
         self._url_prefix = "cinder://"
@@ -466,7 +549,8 @@ class Store(glance_store.driver.Store):
                     for key in ['user_name', 'password',
                                 'project_name', 'auth_address']])
 
-    def get_cinderclient(self, context=None, legacy_update=False):
+    def get_cinderclient(self, context=None, legacy_update=False,
+                         version='3.0'):
         # NOTE: For legacy image update from single store to multiple
         # stores we need to use admin context rather than user provided
         # credentials
@@ -476,15 +560,18 @@ class Store(glance_store.driver.Store):
         else:
             user_overriden = self.is_user_overriden()
 
+        session = get_cinder_session(self.store_conf)
+
         if user_overriden:
             username = self.store_conf.cinder_store_user_name
-            password = self.store_conf.cinder_store_password
-            project = self.store_conf.cinder_store_project_name
             url = self.store_conf.cinder_store_auth_address
+            # use auth that is already in the session
+            auth = None
         else:
             username = context.user_id
-            password = context.auth_token
             project = context.project_id
+            # noauth extracts user_id:project_id from auth_token
+            token = context.auth_token or '%s:%s' % (username, project)
 
             if self.store_conf.cinder_endpoint_template:
                 template = self.store_conf.cinder_endpoint_template
@@ -504,23 +591,19 @@ class Store(glance_store.driver.Store):
                     reason = _("Failed to find Cinder from a service catalog.")
                     raise exceptions.BadStoreConfiguration(store_name="cinder",
                                                            reason=reason)
+            auth = ksa_token_endpoint.Token(endpoint=url, token=token)
 
+        api_version = api_versions.APIVersion(version)
         c = cinderclient.Client(
-            username, password, project, auth_url=url,
+            session=session, auth=auth,
             region_name=self.store_conf.cinder_os_region_name,
-            insecure=self.store_conf.cinder_api_insecure,
             retries=self.store_conf.cinder_http_retries,
-            cacert=self.store_conf.cinder_ca_certificates_file)
+            api_version=api_version)
 
         LOG.debug(
             'Cinderclient connection created for user %(user)s using URL: '
             '%(url)s.', {'user': username, 'url': url})
 
-        # noauth extracts user_id:project_id from auth_token
-        if not user_overriden:
-            c.client.auth_token = context.auth_token or '%s:%s' % (username,
-                                                                   project)
-        c.client.management_url = url
         return c
 
     @contextlib.contextmanager
@@ -612,52 +695,62 @@ class Store(glance_store.driver.Store):
         use_multipath = self.store_conf.cinder_use_multipath
         enforce_multipath = self.store_conf.cinder_enforce_multipath
         mount_point_base = self.store_conf.cinder_mount_point_base
+        volume_id = volume.id
 
-        properties = connector.get_connector_properties(
+        connector_prop = connector.get_connector_properties(
             root_helper, host, use_multipath, enforce_multipath)
 
-        try:
-            volume.reserve(volume)
-        except cinder_exception.ClientException as e:
-            msg = (_('Failed to reserve volume %(volume_id)s: %(error)s')
-                   % {'volume_id': volume.id, 'error': e})
-            LOG.error(msg)
-            raise exceptions.BackendException(msg)
+        if volume.multiattach:
+            attachment = attachment_state_manager.attach(client, volume_id,
+                                                         host,
+                                                         mode=attach_mode)
+        else:
+            attachment = self.volume_api.attachment_create(client, volume_id,
+                                                           mode=attach_mode)
+        attachment = self.volume_api.attachment_update(
+            client, attachment['id'], connector_prop,
+            mountpoint='glance_store')
+        self.volume_api.attachment_complete(client, attachment.id)
+        volume = volume.manager.get(volume_id)
+        connection_info = attachment.connection_info
 
         try:
-            connection_info = volume.initialize_connection(volume, properties)
             conn = connector.InitiatorConnector.factory(
                 connection_info['driver_volume_type'], root_helper,
-                conn=connection_info)
+                conn=connection_info, use_multipath=use_multipath)
             if connection_info['driver_volume_type'] == 'nfs':
-                if volume.encrypted:
-                    volume.unreserve(volume)
-                    volume.delete()
-                    msg = (_('Encrypted volume creation for cinder nfs is not '
-                             'supported from glance_store. Failed to create '
-                             'volume %(volume_id)s')
-                           % {'volume_id': volume.id})
+                # The format info of nfs volumes is exposed via attachment_get
+                # API hence it is not available in the connection info of
+                # attachment object received from attachment_update and we
+                # need to do this call
+                vol_attachment = self.volume_api.attachment_get(
+                    client, attachment.id)
+                if (volume.encrypted or
+                        vol_attachment.connection_info['format'] == 'qcow2'):
+                    issue_type = 'Encrypted' if volume.encrypted else 'qcow2'
+                    msg = (_('%(issue_type)s volume creation for cinder nfs '
+                             'is not supported from glance_store. Failed to '
+                             'create volume %(volume_id)s')
+                           % {'issue_type': issue_type,
+                              'volume_id': volume_id})
                     LOG.error(msg)
                     raise exceptions.BackendException(msg)
 
-                @utils.synchronized(connection_info['data']['export'])
+                @utils.synchronized(connection_info['export'])
                 def connect_volume_nfs():
-                    data = connection_info['data']
-                    export = data['export']
-                    vol_name = data['name']
+                    export = connection_info['export']
+                    vol_name = connection_info['name']
                     mountpoint = self._get_mount_path(
                         export,
                         os.path.join(mount_point_base, 'nfs'))
-                    options = data['options']
+                    options = connection_info['options']
                     self.mount.mount(
                         'nfs', export, vol_name, mountpoint, host,
                         root_helper, options)
                     return {'path': os.path.join(mountpoint, vol_name)}
                 device = connect_volume_nfs()
             else:
-                device = conn.connect_volume(connection_info['data'])
-            volume.attach(None, 'glance_store', attach_mode, host_name=host)
-            volume = self._wait_volume_status(volume, 'attaching', 'in-use')
+                device = conn.connect_volume(connection_info)
             if (connection_info['driver_volume_type'] == 'rbd' and
                not conn.do_local_attach):
                 yield device['path']
@@ -670,38 +763,29 @@ class Store(glance_store.driver.Store):
                               '%(volume_id)s.'), {'volume_id': volume.id})
             raise
         finally:
-            if volume.status == 'in-use':
-                volume.begin_detaching(volume)
-            elif volume.status == 'attaching':
-                volume.unreserve(volume)
-
             if device:
                 try:
                     if connection_info['driver_volume_type'] == 'nfs':
-                        @utils.synchronized(connection_info['data']['export'])
+                        @utils.synchronized(connection_info['export'])
                         def disconnect_volume_nfs():
                             path, vol_name = device['path'].rsplit('/', 1)
                             self.mount.umount(vol_name, path, host,
                                               root_helper)
                         disconnect_volume_nfs()
                     else:
-                        conn.disconnect_volume(connection_info['data'], device)
+                        if volume.multiattach:
+                            attachment_state_manager.detach(
+                                client, attachment.id, volume_id, host, conn,
+                                connection_info, device)
+                        else:
+                            conn.disconnect_volume(connection_info, device)
                 except Exception:
                     LOG.exception(_LE('Failed to disconnect volume '
                                       '%(volume_id)s.'),
                                   {'volume_id': volume.id})
 
-            try:
-                volume.terminate_connection(volume, properties)
-            except Exception:
-                LOG.exception(_LE('Failed to terminate connection of volume '
-                                  '%(volume_id)s.'), {'volume_id': volume.id})
-
-            try:
-                client.volumes.detach(volume)
-            except Exception:
-                LOG.exception(_LE('Failed to detach volume %(volume_id)s.'),
-                              {'volume_id': volume.id})
+            if not volume.multiattach:
+                self.volume_api.attachment_delete(client, attachment.id)
 
     def _cinder_volume_data_iterator(self, client, volume, max_size, offset=0,
                                      chunk_size=None, partial_length=None):
@@ -748,7 +832,7 @@ class Store(glance_store.driver.Store):
         loc = location.store_location
         self._check_context(context)
         try:
-            client = self.get_cinderclient(context)
+            client = self.get_cinderclient(context, version='3.54')
             volume = client.volumes.get(loc.volume_id)
             size = int(volume.metadata.get('image_size',
                                            volume.size * units.Gi))
@@ -816,7 +900,7 @@ class Store(glance_store.driver.Store):
         """
 
         self._check_context(context, require_tenant=True)
-        client = self.get_cinderclient(context)
+        client = self.get_cinderclient(context, version='3.54')
         os_hash_value = utils.get_hasher(hashing_algo, False)
         checksum = utils.get_hasher('md5', False)
         bytes_written = 0
@@ -838,9 +922,9 @@ class Store(glance_store.driver.Store):
                          "resize-before-write for each GB which "
                          "will be considerably slower than normal."))
         try:
-            volume = client.volumes.create(size_gb, name=name,
-                                           metadata=metadata,
-                                           volume_type=volume_type)
+            volume = self.volume_api.create(client, size_gb, name=name,
+                                            metadata=metadata,
+                                            volume_type=volume_type)
         except cinder_exception.NotFound:
             LOG.error(_LE("Invalid volume type %s configured. Please check "
                           "the `cinder_volume_type` configuration parameter."
@@ -949,9 +1033,9 @@ class Store(glance_store.driver.Store):
         """
         loc = location.store_location
         self._check_context(context)
+        client = self.get_cinderclient(context)
         try:
-            volume = self.get_cinderclient(context).volumes.get(loc.volume_id)
-            volume.delete()
+            self.volume_api.delete(client, loc.volume_id)
         except cinder_exception.NotFound:
             raise exceptions.NotFound(image=loc.volume_id)
         except cinder_exception.ClientException as e:
