@@ -26,12 +26,13 @@ import tempfile
 import time
 import uuid
 
-from cinderclient.v3 import client as cinderclient
 from os_brick.initiator import connector
 from oslo_concurrency import processutils
 from oslo_utils.secretutils import md5
 from oslo_utils import units
 
+from glance_store.common import attachment_state_manager
+from glance_store.common import cinder_utils
 from glance_store import exceptions
 from glance_store import location
 from glance_store.tests import base
@@ -67,31 +68,30 @@ class TestCinderStore(base.StoreBaseTest,
                                   auth_token='fake_token',
                                   project_id='fake_project')
         self.hash_algo = 'sha256'
+        cinder._reset_cinder_session()
 
     def test_get_cinderclient(self):
         cc = self.store.get_cinderclient(self.context)
-        self.assertEqual('fake_token', cc.client.auth_token)
-        self.assertEqual('http://foo/public_url', cc.client.management_url)
+        self.assertEqual('fake_token', cc.client.auth.token)
+        self.assertEqual('http://foo/public_url', cc.client.auth.endpoint)
 
-    def test_get_cinderclient_with_user_overriden(self):
+    def _test_get_cinderclient_with_user_overriden(self):
         self.config(cinder_store_user_name='test_user')
         self.config(cinder_store_password='test_password')
         self.config(cinder_store_project_name='test_project')
         self.config(cinder_store_auth_address='test_address')
         cc = self.store.get_cinderclient(self.context)
-        self.assertIsNone(cc.client.auth_token)
-        self.assertEqual('test_address', cc.client.management_url)
+        self.assertEqual('test_project', cc.client.session.auth.project_name)
+        self.assertEqual('Default', cc.client.session.auth.project_domain_name)
+        return cc
+
+    def test_get_cinderclient_with_user_overriden(self):
+        self._test_get_cinderclient_with_user_overriden()
 
     def test_get_cinderclient_with_user_overriden_and_region(self):
         self.config(cinder_os_region_name='test_region')
-        fake_client = FakeObject(client=FakeObject(auth_token=None))
-        with mock.patch.object(cinderclient, 'Client',
-                               return_value=fake_client) as mock_client:
-            self.test_get_cinderclient_with_user_overriden()
-            mock_client.assert_called_once_with(
-                'test_user', 'test_password', 'test_project',
-                auth_url='test_address', cacert=None, insecure=False,
-                region_name='test_region', retries=3)
+        cc = self._test_get_cinderclient_with_user_overriden()
+        self.assertEqual('test_region', cc.client.region_name)
 
     def test_temporary_chown(self):
         class fake_stat(object):
@@ -153,11 +153,22 @@ class TestCinderStore(base.StoreBaseTest,
     def _test_open_cinder_volume(self, open_mode, attach_mode, error,
                                  multipath_supported=False,
                                  enforce_multipath=False,
-                                 encrypted_nfs=False):
+                                 encrypted_nfs=False, qcow2_vol=False,
+                                 multiattach=False):
         self.config(cinder_mount_point_base=None)
-        fake_volume = mock.MagicMock(id=str(uuid.uuid4()), status='available')
-        fake_volumes = FakeObject(get=lambda id: fake_volume,
-                                  detach=mock.Mock())
+        fake_volume = mock.MagicMock(id=str(uuid.uuid4()), status='available',
+                                     multiattach=multiattach)
+        fake_volume.manager.get.return_value = fake_volume
+        fake_volumes = FakeObject(get=lambda id: fake_volume)
+        fake_attachment_id = str(uuid.uuid4())
+        fake_attachment_create = {'id': fake_attachment_id}
+        if encrypted_nfs or qcow2_vol:
+            fake_attachment_update = mock.MagicMock(
+                id=fake_attachment_id,
+                connection_info={'driver_volume_type': 'nfs'})
+        else:
+            fake_attachment_update = mock.MagicMock(id=fake_attachment_id)
+        fake_conn_info = mock.MagicMock(connector={})
         fake_client = FakeObject(volumes=fake_volumes)
         _, fake_dev_path = tempfile.mkstemp(dir=self.test_dir)
         fake_devinfo = {'path': fake_dev_path}
@@ -170,14 +181,22 @@ class TestCinderStore(base.StoreBaseTest,
             yield
 
         def do_open():
-            with self.store._open_cinder_volume(
-                    fake_client, fake_volume, open_mode):
-                if error:
-                    raise error
+            if multiattach:
+                with mock.patch.object(
+                        attachment_state_manager._AttachmentStateManager,
+                        'get_state') as mock_get_state:
+                    mock_get_state.return_value.__enter__.return_value = (
+                        attachment_state_manager._AttachmentState())
+                    with self.store._open_cinder_volume(
+                            fake_client, fake_volume, open_mode):
+                        pass
+            else:
+                with self.store._open_cinder_volume(
+                        fake_client, fake_volume, open_mode):
+                    if error:
+                        raise error
 
         def fake_factory(protocol, root_helper, **kwargs):
-            self.assertEqual(fake_volume.initialize_connection.return_value,
-                             kwargs['conn'])
             return fake_connector
 
         root_helper = "sudo glance-rootwrap /etc/glance/rootwrap.conf"
@@ -189,30 +208,44 @@ class TestCinderStore(base.StoreBaseTest,
                 mock.patch.object(cinder.Store, 'get_root_helper',
                                   return_value=root_helper), \
                 mock.patch.object(connector.InitiatorConnector, 'factory',
-                                  side_effect=fake_factory):
+                                  side_effect=fake_factory
+                                  ) as fake_conn_obj, \
+                mock.patch.object(cinder_utils.API, 'attachment_create',
+                                  return_value=fake_attachment_create
+                                  ) as attach_create, \
+                mock.patch.object(cinder_utils.API, 'attachment_update',
+                                  return_value=fake_attachment_update
+                                  ) as attach_update, \
+                mock.patch.object(cinder_utils.API,
+                                  'attachment_delete') as attach_delete, \
+                mock.patch.object(cinder_utils.API,
+                                  'attachment_get') as attach_get, \
+                mock.patch.object(cinder_utils.API,
+                                  'attachment_complete') as attach_complete:
 
             with mock.patch.object(connector,
-                                   'get_connector_properties') as mock_conn:
+                                   'get_connector_properties',
+                                   return_value=fake_conn_info) as mock_conn:
                 if error:
                     self.assertRaises(error, do_open)
-                elif encrypted_nfs:
-                    fake_volume.initialize_connection.return_value = {
-                        'driver_volume_type': 'nfs'
-                    }
-                    fake_volume.encrypted = True
+                elif encrypted_nfs or qcow2_vol:
+                    fake_volume.encrypted = False
+                    if encrypted_nfs:
+                        fake_volume.encrypted = True
+                    elif qcow2_vol:
+                        attach_get.return_value = mock.MagicMock(
+                            connection_info={'format': 'qcow2'})
                     try:
                         with self.store._open_cinder_volume(
                                 fake_client, fake_volume, open_mode):
                             pass
                     except exceptions.BackendException:
-                        self.assertEqual(1,
-                                         fake_volume.unreserve.call_count)
-                        self.assertEqual(1,
-                                         fake_volume.delete.call_count)
+                        attach_delete.assert_called_once_with(
+                            fake_client, fake_attachment_id)
                 else:
                     do_open()
 
-                if not encrypted_nfs:
+                if not (encrypted_nfs or qcow2_vol):
                     mock_conn.assert_called_once_with(
                         root_helper, socket.gethostname(),
                         multipath_supported, enforce_multipath)
@@ -220,10 +253,36 @@ class TestCinderStore(base.StoreBaseTest,
                         mock.ANY)
                     fake_connector.disconnect_volume.assert_called_once_with(
                         mock.ANY, fake_devinfo)
-                    fake_volume.attach.assert_called_once_with(
-                        None, 'glance_store', attach_mode,
-                        host_name=socket.gethostname())
-                    fake_volumes.detach.assert_called_once_with(fake_volume)
+                    fake_conn_obj.assert_called_once_with(
+                        mock.ANY, root_helper, conn=mock.ANY,
+                        use_multipath=multipath_supported)
+                    attach_create.assert_called_once_with(
+                        fake_client, fake_volume.id, mode=attach_mode)
+                    attach_update.assert_called_once_with(
+                        fake_client, fake_attachment_id,
+                        fake_conn_info, mountpoint='glance_store')
+                    attach_complete.assert_called_once_with(
+                        fake_client, fake_attachment_id)
+                    attach_delete.assert_called_once_with(
+                        fake_client, fake_attachment_id)
+                else:
+                    mock_conn.assert_called_once_with(
+                        root_helper, socket.gethostname(),
+                        multipath_supported, enforce_multipath)
+                    fake_connector.connect_volume.assert_not_called()
+                    fake_connector.disconnect_volume.assert_not_called()
+                    fake_conn_obj.assert_called_once_with(
+                        mock.ANY, root_helper, conn=mock.ANY,
+                        use_multipath=multipath_supported)
+                    attach_create.assert_called_once_with(
+                        fake_client, fake_volume.id, mode=attach_mode)
+                    attach_update.assert_called_once_with(
+                        fake_client, fake_attachment_id,
+                        fake_conn_info, mountpoint='glance_store')
+                    attach_complete.assert_called_once_with(
+                        fake_client, fake_attachment_id)
+                    attach_delete.assert_called_once_with(
+                        fake_client, fake_attachment_id)
 
     def test_open_cinder_volume_rw(self):
         self._test_open_cinder_volume('wb', 'rw', None)
@@ -249,6 +308,12 @@ class TestCinderStore(base.StoreBaseTest,
     def test_open_cinder_volume_nfs_encrypted(self):
         self._test_open_cinder_volume('rb', 'ro', None, encrypted_nfs=True)
 
+    def test_open_cinder_volume_nfs_qcow2_volume(self):
+        self._test_open_cinder_volume('rb', 'ro', None, qcow2_vol=True)
+
+    def test_open_cinder_volume_multiattach_volume(self):
+        self._test_open_cinder_volume('rb', 'ro', None, multiattach=True)
+
     def test_cinder_configure_add(self):
         self.assertRaises(exceptions.BadStoreConfiguration,
                           self.store._check_context, None)
@@ -399,8 +464,7 @@ class TestCinderStore(base.StoreBaseTest,
     def test_cinder_delete(self):
         fake_client = FakeObject(auth_token=None, management_url=None)
         fake_volume_uuid = str(uuid.uuid4())
-        fake_volume = FakeObject(delete=mock.Mock())
-        fake_volumes = {fake_volume_uuid: fake_volume}
+        fake_volumes = FakeObject(delete=mock.Mock())
 
         with mock.patch.object(cinder.Store, 'get_cinderclient') as mocked_cc:
             mocked_cc.return_value = FakeObject(client=fake_client,
@@ -409,7 +473,7 @@ class TestCinderStore(base.StoreBaseTest,
             uri = 'cinder://%s' % fake_volume_uuid
             loc = location.get_location_from_uri(uri, conf=self.conf)
             self.store.delete(loc, context=self.context)
-            fake_volume.delete.assert_called_once_with()
+            fake_volumes.delete.assert_called_once_with(fake_volume_uuid)
 
     def test_set_url_prefix(self):
         self.assertEqual('cinder://', self.store._url_prefix)
